@@ -1,5 +1,5 @@
 /*
-Copyright (C) 2000-2005  The PARI group.
+Copyright (C) 2000-2013  The PARI group.
 
 This file is part of the GP2C package.
 
@@ -10,7 +10,7 @@ ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
 with the package; see the file 'COPYING'. If not, write to the Free Software
-Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */
+Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.*/
 
 #include "config.h"
 #include <stdlib.h>
@@ -19,16 +19,36 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */
 #include "header.h"
 #include "patchfunc.h"
 extern int indent;
+
+int has_iter(char *s)
+{
+  int nf=getfunc(s);
+  return !!lfunc[nf].iter;
+}
+
+void patchfunction(struct patchfunc_s *p)
+{
+  int nf=getfunc(p->gpname);
+  lfunc[nf].proto.code=p->code;
+  functype(lfunc[nf])=p->type;
+  funcmode(lfunc[nf])=p->mode;
+  lfunc[nf].spec=p->spec;
+}
+
 void patchfunclist(void)
 {
   int i;
   for(i=0;patchfunc[i].gpname;i++)
+    patchfunction(patchfunc+i);
+  if (!has_iter("forvec"))
   {
-    int nf=getfunc(patchfunc[i].gpname);
-    lfunc[nf].proto.code=patchfunc[i].code;
-    functype(lfunc[nf])=patchfunc[i].type;
-    funcmode(lfunc[nf])=patchfunc[i].mode;
-    lfunc[nf].spec=patchfunc[i].spec;
+    struct patchfunc_s pforvec = {"forvec",NULL,Gnotype,(1<<Msemicolon),GPforvec};
+    patchfunction(&pforvec);
+  }
+  if (!has_iter("forprime"))
+  {
+    struct patchfunc_s pforprime = {"forprime",NULL,Gnotype,(1<<Msemicolon),GPforprime};
+    patchfunction(&pforprime);
   }
 }
 void printlistfunc(FILE *fout)
@@ -82,12 +102,15 @@ int guesstype(int n)
       return Gstr;
     }
   case Fentry:
-    return vartype(ctxstack[getvarerr(n)]);
+    {
+      long v = getvar(n);
+      return v<0 ? Ggen: vartype(ctxstack[v]);
+    }
   case Ffunction:
     if (!strcmp(entryname(n),"length"))
       return Glg;
   default: /*FALL THROUGH*/
-    return Gnotype;
+    return Ggen;
   }
 }
 
@@ -113,17 +136,30 @@ void genequaltosmall(int binf,int bsup, int *tv, int def)
 
 int gengerepile(int bl)
 {
-  int arg[4];
+  int arg[3];
   block[bl].gc|=(1<<GCneeded);
   arg[0]=newnode(Fentry,newentry("btop"),-1);
   pushvar(arg[0],0,Gpari_sp,newcall("_avma",-1));
-  arg[1]=newnode(Fentry,newentry("st_lim"),-1);
-  pushvar(arg[1],0,Gpari_sp,
-      newcall("_stack_lim",newnode(Flistarg,arg[0],newsmall(1))));
-  arg[2]=newsmall(bl);
+  arg[1]=newsmall(bl);
+  if (FC_gc_needed>=0)
+    arg[2]=GNOARG;
+  else
+  {
+    arg[2]=newnode(Fentry,newentry("st_lim"),-1);
+    pushvar(arg[2],0,Gpari_sp,
+        newcall("_stack_lim",newnode(Flistarg,arg[0],newsmall(1))));
+  }
   return stacktoargs(arg,3);
 }
 
+int gengerepilereturn(int bl)
+{
+  int arg[2];
+  arg[0]=newnode(Fentry,newentry("ltop"),-1);
+  arg[1]=newsmall(bl);
+  return stacktoargs(arg,2);
+}
+
 int newgetg(int v, int l, const char *t)
 {
   return newcall("_cgetg", newnode(Flistarg, newnode(Flistarg,
@@ -133,8 +169,10 @@ int newgetg(int v, int l, const char *t)
 void makesubblock_ctx(int n, int savx)
 {
   int bl3=newblock();
+  int level = preclevel;
   genblock(n,-1);
   makeblock(bl3,n,newleaf(n),-1,savx);
+  preclevel = level;
 }
 
 void makesubblock(int n)
@@ -189,6 +227,127 @@ int get_iferr(int k)
   return 0;
 }
 
+int
+geniterproto(int n, long fl, gpfunc *gp, int *argo, int *seq, int *var, int *tv)
+{
+  int arg[STACKSZ];
+  const char *p=gp->proto.code;
+  char c;
+  PPproto mod;
+  int nb=genlistargs(n,arg,0,STACKSZ);
+  int i=0, k=0, binf;
+  while((mod=parseproto(&p,&c))!=PPend)
+  {
+    if (i<nb && tree[arg[i]].f!=Fnoarg && (mod==PPdefault || mod==PPdefaultmulti))
+      mod=PPstd;
+    switch(mod)
+    {
+    case PPstd:
+      if (i>=nb)
+        die(n,"too few arguments");
+      switch(c)
+      {
+      case '=':
+        genequal(arg[i], gp->gpname, var, &binf, tv);
+        if (fl)
+        {
+          tree[*var].t = *tv;
+          argo[k++] = *var;
+          tree[binf].t = guesstype(binf);
+        }
+        argo[k++] = binf;
+        break;
+      case 'V':
+        checkisvar(arg[i], gp->gpname, var, tv);
+        if (fl)
+        {
+          tree[*var].t = *tv;
+          argo[k++] = *var;
+        }
+        break;
+      case 'I':
+        *seq = arg[i];
+        break;
+      default:
+        if (fl)
+          tree[arg[i]].t = guesstype(arg[i]);
+        argo[k++] = arg[i];
+      }
+      break;
+    case PPdefault:
+    case PPdefaultmulti:
+      if (i < nb)
+      {
+        tree[arg[i]].t = guesstype(arg[i]);
+        argo[k++] = arg[i];
+      }
+      break;
+    case PPstar:
+      i=nb-1;
+      break;
+    case PPauto: /* Skip */
+      i--;
+      break;
+    default:
+      break;
+    }
+    i++;
+  }
+  return k;
+}
+
+void
+genblockfunciter(int n, gpfunc *gp)
+{
+  int arg[STACKSZ];
+  int var=-1, tv=Gnotype;
+  int bseq=-1,aseq=-1,ret=-1,vprime,seq,cseq;
+  int gpm=-1,bl0,bl1,bl2;
+  int sav0, savx;
+  Clabel_t lab;
+  gpiterator * iter;
+  const char * init, *next;
+  int nb = geniterproto(n, 1, gp, arg, &seq, &var, &tv);
+  gpdescarg * da = descfindrulesdsc(nb, arg, gp->iter);
+  if (!da) die(n, "no suitable iterator found");
+  iter = &(da->iter);
+  init = lfunc[iter->init].gpname;
+  next = lfunc[iter->next].gpname;
+  sav0=s_ctx.n;
+  bl0=newblock();
+  vprime=newnode(Fentry,newentry("iter"),-1);
+  pushvar(vprime,1<<Cuser,iter->type,-1);
+  nb = geniterproto(n, 0, gp, arg+1, &seq, &var, &tv)+1;
+  arg[0] = vprime;
+  bseq = newcall(init, stacktoargs(arg,nb));
+  genblock(bseq,-1);
+  savx=s_ctx.n;
+  bl1=newblock();
+  bl2=newblock();
+  if (autogc)
+    gpm=gengerepile(bl2);
+  if (tv==Gnotype)
+  {
+    tree[vprime].t = iter->type;
+    gpdescarg *da=descfindrules(1, &vprime, lfunc+iter->next);
+    tv = da->type;
+  }
+  pushvar(var,1<<Cuser|1<<Ccompo,tv,-1);
+  lab=makesubblock_label(seq,Lbreak);
+  if (lab.ne)
+    seq=newseq(seq,newcall("_label",newsmall(lab.ne)));
+  if (autogc)
+    seq=addseqright(seq,newcall("_gerepilemany",gpm));
+  cseq=newnode(Fassign,var,newcall(next,vprime));
+  tree[cseq].m=Mparens;
+  aseq=addseqright(aseq,newcall("_whilenull",newnode(Flistarg,cseq,seq)));
+  if (lab.go)
+    aseq=newseq(aseq,newcall("_label",newsmall(lab.go)));
+  makeblocks3(bl0,bl1,bl2,n,bseq,aseq,ret,savx,savx);
+  copyctx(sav0,block+bl0);
+  restorectx(sav0);
+}
+
 /*
    The 6 operations which order is very important are
 are:
@@ -244,8 +403,8 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
     break;
   case GPif:
     nb=genlistargs(n,arg,2,STACKSZ);
-    /*Some people type 'if(x!=0,print(x),)'*/
-    while (nb>=3 && arg[nb-1]==GNOARG)
+    if (nb==3 && arg[2]==GNOARG)
+      /*Some people type 'if(x!=0,print(x),)'*/
       nb--;
     if (nb==3 && arg[1]==GNOARG)
     {
@@ -265,29 +424,33 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
     }
     else
     {
-      int i;
       if (p>=0 && tree[p].f!=Fseq)
         newdecl((1<<Cauto)|(1<<Cimmutable),Gvoid,GNIL,&ret);
+      if (nb>=4)
+      {
+        arg[2] = newcall("if", stacktoargs(arg+2,nb-2));
+        nb = 3;
+      }
+      if (arg[0]>=0 && tree[arg[0]].f==Fseq)
+      {
+        int pred;
+        newdecl((1<<Cauto)|(1<<Cimmutable),Gbool,GNIL,&pred);
+        aseq = geninsertvar(arg[0], pred);
+        genblock(aseq,n);
+        arg[0] = newleaf(pred);
+      }
       genblock(arg[0],n);
-      arg[1]=geninsertvar(arg[1],ret);
-      makesubblock(arg[1]);
-      for(i=2; i<nb-1; i+=2)
+      if (arg[1]!=GNOARG)
       {
-        if (arg[i] == GNOARG)
-          arg[i] = GNIL;
-        else
-          genblock(arg[i],n);
-        if (ret>0)
-          arg[i+1]=geninsertvar(arg[i+1],newleaf(ret));
-        makesubblock(arg[i+1]);
+        arg[1]=geninsertvar(arg[1],ret);
+        makesubblock(arg[1]);
       }
-      if (i<nb && arg[i]>=0)
+      if(nb==3)
       {
-        if (ret>0)
-          arg[i]=geninsertvar(arg[i],newleaf(ret));
-        makesubblock(arg[i]);
+        arg[2]=geninsertvar(arg[2],ret==-1?-1:newleaf(ret));
+        makesubblock(arg[2]);
       }
-      aseq=newnode(tree[n].f,tree[n].x,stacktoargs(arg,nb));
+      aseq=addseqright(aseq,newnode(tree[n].f,tree[n].x,stacktoargs(arg,nb)));
     }
     makeblock(newblock(),n,aseq,ret,s_ctx.n);
     break;
@@ -335,6 +498,9 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       case 'l':
         functype(*gp)=Gsmall;
         break;
+      case 'u':
+        functype(*gp)=Gusmall;
+        break;
       case 'i':
         functype(*gp)=Gsmall_int;
         break;
@@ -348,7 +514,7 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       }
       if (tree[arg[1]].f==Ftag)
         functype(*gp)=tree[arg[1]].y;
-      for(;*p;p++) if (*p=='p') funcmode(*gp)|=(1<<Mprec);
+      for(;*p;p++) if (*p=='p' || *p=='b') funcmode(*gp)|=(1<<Mprec);
       gp->spec=GPinstalled;
     }
     break;
@@ -466,7 +632,7 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
     }
   case GPforstep:
     {
-      int bstep;
+      int bstep,lstep,sstep;
       int vstep,vindex,vsign;
       int dosign,isvec;
       nb=genlistargs(n,arg,4,4);
@@ -480,14 +646,34 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       bseq=addseqright(bseq,geninsertvar(bstep,newleaf(vstep)));
       if (guesstype(bstep)==Gvec)
       {
-        isvec=1;
-        newdecl(0,Gsmall,-1,&vindex);
-        dosign=0;
+        int sarg[3];
+        isvec=1; dosign=1;
+        newdecl((1<<Cauto)|(1<<Cconst),Gsmall,-1,&vindex);
+        newdecl((1<<Cauto)|(1<<Cconst),Gsmall,-1,&lstep);
+        bseq=addseqright(bseq,geninsertvar(newcall("length",newleaf(vstep)),newleaf(lstep)));
+        sarg[0] = newnode(Fassign,newleaf(vindex),newsmall(1));
+        sarg[1] = newleaf(lstep);
+        sarg[2] = newnode(Fmatcoeff,newleaf(vstep),
+            newnode(Fmatrix,newnode(Frange,newleaf(vindex),GNOARG),-1));
+        sstep = newcall("sum", stacktoargs(sarg,3));
+        if (tv==Gnotype) tv=Ggen;
       }
       else
       {
         isvec=0;
         dosign=tree[bstep].f!=Fsmall;
+        if (dosign)
+        {
+          sstep = newleaf(vstep);
+          if (tv==Gnotype) tv=Ggen;
+        } else
+          genequaltosmall(binf,bsup,&tv,Ggen);
+      }
+      if (dosign)
+      {
+        newdecl((1<<Cauto)|(1<<Cconst),Gbool,-1,&vsign);
+        sstep = newopcall(OPg,sstep, newsmall(0));
+        bseq=addseqright(bseq,geninsertvar(sstep,newleaf(vsign)));
       }
       genblock(bseq,-1);
       savx=s_ctx.n;
@@ -495,44 +681,37 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       bl2=newblock();
       if (autogc)
         gpm=gengerepile(bl2);
-      if (tv==Gnotype)
-        tv=Ggen;
       pushvar(var,1<<Cuser,tv,-1);
-      if (dosign)
-        newdecl(0,Gbool,newopcall(OPg,newleaf(vstep),0),&vsign);
       lab=makesubblock_label(seq,Lbreak);
       if (lab.ne)
         seq=newseq(seq,newcall("_label",newsmall(lab.ne)));
       if (autogc)
         seq=addseqright(seq,newcall("_gerepilemany",gpm));
       arg[0]=newnode(Fassign,var,vinf);
-      if (isvec)
+      if ( !dosign )
       {
-        arg[4]=vindex;
-        arg[5]=newcall("length",vstep);
-        arg[3]=newopcall(OPle,newleaf(var),vsup);
-        arg[2]=newopcall(OPpe,var,newcoeff(vstep,vindex,-1));
-        nb=6;
+        if ( tree[bstep].x>0 )
+          arg[3]=newopcall(OPle,newleaf(var),vsup);
+        else
+          arg[3]=newopcall(OPge,newleaf(var),vsup);
+        nb=4;
       }
       else
       {
-        if ( !dosign )
-        {
-          if ( tree[bstep].x>0 )
-            arg[3]=newopcall(OPle,newleaf(var),vsup);
-          else
-            arg[3]=newopcall(OPge,newleaf(var),vsup);
-          nb=4;
-        }
-        else
-        {
-          arg[3]=newleaf(vsign);
-          arg[4]=newopcall(OPle,newleaf(var),vsup);
-          arg[5]=newopcall(OPge,newleaf(var),vsup);
-          nb=6;
-        }
-        arg[2]=newopcall(OPpe,var,vstep);
+        arg[3]=newleaf(vsign);
+        arg[4]=newopcall(OPle,newleaf(var),vsup);
+        arg[5]=newopcall(OPge,newleaf(var),vsup);
+        nb=6;
       }
+      if (isvec)
+      {
+        arg[2]=newopcall(OPpe,var,newcoeff(vstep,vindex,-1));
+        arg[6]=vindex;
+        arg[7]=newleaf(lstep);
+        nb=8;
+      }
+      else
+        arg[2]=newopcall(OPpe,var,vstep);
       arg[1]=seq;
       if (isvec)
         aseq=newcall("_forstepvec",stacktoargs(arg,nb));
@@ -543,8 +722,20 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       makeblocks(bl1,bl2,n,bseq,aseq,ret,savx);
     }
     break;
-  case GPforprime:
   case GPprodeuler:
+    nb=genlistargs(n,arg,3,3);
+    seq = arg[2];
+    newdecl(0,Ggen,-1,&ret);
+    savx=s_ctx.n;
+    bl1=newblock();
+    arg[2]=geninsertvarop(seq,ret,OPme);
+    aseq=newnode(Fseq,
+            geninsertvar(newnode(Fconst,newsmallrealvalue(1),-1),newleaf(ret)),
+                 newcall("forprime",stacktoargs(arg,3)));
+    genblock(aseq,-1);
+    makeblock(bl1,n,aseq,ret,savx);
+    break;
+  case GPforprime:
     nb=genlistargs(n,arg,3,3);
     genequal(arg[0],gp->gpname,&var,&binf,&tv);
     bsup=arg[1];seq=arg[2];
@@ -555,12 +746,6 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       newdecl((1<<Cauto)|(1<<Cconst),Gempty,-1,&vsup);
       bseq=addseqright(bseq,geninsertvar(bsup,newleaf(vsup)));
     }
-    if (gp->spec==GPprodeuler)
-    {
-      newdecl(0,Ggen,-1,&ret);
-      aseq=addseqright(aseq,geninsertvar(
-            newnode(Fconst,newsmallrealvalue(1),-1),newleaf(ret)));
-    }
     genblock(bseq,-1);
     savx=s_ctx.n;
     bl1=newblock();
@@ -571,18 +756,9 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
     pushvar(var,1<<Cuser,tv,FC_forprime_init>=0?-1:newsmall(0));
     vprime=newnode(Fentry,newentry("primepointer"),-1);
     pushvar(vprime,1<<Cuser,Gforprime,FC_forprime_init>=0?-1:newcall("_diffptr",-1));
-    if (gp->spec==GPprodeuler)
-    {
-      seq=geninsertvarop(seq,ret,OPme);
-      makesubblock_label(seq,Lnobrk);
-      lab.go=0;
-    }
-    else
-    {
-      lab=makesubblock_label(seq,Lbreak);
-      if (lab.ne)
-        seq=newseq(seq,newcall("_label",newsmall(lab.ne)));
-    }
+    lab=makesubblock_label(seq,Lbreak);
+    if (lab.ne)
+      seq=newseq(seq,newcall("_label",newsmall(lab.ne)));
     if (autogc)
       seq=addseqright(seq,newcall("_gerepilemany",gpm));
     if (FC_forprime_init>=0)
@@ -619,7 +795,7 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
               newleaf(var),newleaf(vprime))),seq);
       aseq=addseqright(aseq,newcall("_doloop",seq));
     }
-    if (gp->spec==GPforprime && lab.go)
+    if (lab.go)
       aseq=newseq(aseq,newcall("_label",newsmall(lab.go)));
     makeblocks(bl1,bl2,n,bseq,aseq,ret,savx);
     break;
@@ -801,13 +977,44 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       int flag=1<<Cuser;
       int mint=(tree[p].f==Ftag)?tree[p].y:-1;
       if (gp->spec==GPglobal)
-        flag|=1<<Cglobal;
+        flag|=(1<<Cglobal)|(1<<Cfunction);
       savx=s_ctx.n;
       (void)genblockdeclaration(y,n,flag,mint,&aseq);
       makeblock(newblock(),n,aseq,-1,s_ctx.n);
       if (tree[p].f==Ftag) tree[p]=tree[n];
     }
     break;
+  case GPself:
+    {
+      int str = newstringnode(lfunc[currfunc].gpname,-1);
+      int aseq = newcall("_strtoclosure",str);
+      tree[n] = tree[aseq];
+    }
+    break;
+  case GPlocalprec:
+    {
+      char buf[64];
+      int vn;
+      sprintf(buf,"prec%d",++preclevel);
+      vn = newnode(Fentry, newentry(strdup(buf)), -1);
+      pushvar(vn, (1<<Cuser), Gsmall, -1);
+      y = newnode(Fassign, vn, newcall("_ndec2prec", y));
+      genblock(y, n);
+      tree[n].y = y;
+    }
+    break;
+  case GPlocalbitprec:
+    {
+      char buf[64];
+      int vn;
+      sprintf(buf,"bitprec%d",++preclevel);
+      vn = newnode(Fentry, newentry(strdup(buf)), -1);
+      pushvar(vn, (1<<Cuser), Gsmall, -1);
+      y = newnode(Fassign, vn, newcall("_ndec2nbits", y));
+      genblock(y, n);
+      tree[n].y = y;
+    }
+    break;
   case GPreturn:
     {
       int iferr_reset=0;
@@ -845,10 +1052,7 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       block[bl2].egc=var;
       if (autogc)
       {
-        arg[0]=newnode(Fentry,newentry("ltop"),-1);
-        arg[1]=GNOARG;
-        arg[2]=newsmall(bl2);
-        gpm=stacktoargs(arg,3);
+        gpm=gengerepilereturn(bl2);
         aseq=addseqright(aseq,newcall("_gerepilemany",gpm));
       }
       if (iferr_reset)
@@ -905,9 +1109,15 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       checkisvar(arg[1],gp->gpname,&var,&tv);
       seq=arg[0];
       bseq=arg[2];
-      if (tv==Gnotype) tv=Gerror;
-      newdecl((1<<Cauto)|(1<<Cimmutable),Gvoid,GNIL,&ret);
+      if (tv==Gnotype)
+      {
+        tv=Gerror;
+        if (tv<0) die(n,"iferr not supported by your GP version");
+      }
+      if (p>=0 && tree[p].f!=Fseq)
+        newdecl((1<<Cauto)|(1<<Cimmutable),Gvoid,GNIL,&ret);
       seq=geninsertvar(seq,ret);
+      bseq=geninsertvar(bseq,ret);
       lab=makesubblock_label(seq,Liferr);
       savx=s_ctx.n;
       pushvar(var,1<<Cuser,tv,newcall("_iferr_error",-1));
@@ -952,23 +1162,25 @@ int gentypefuncspec(int n, gpfunc *gp)
   switch(gp->spec)
   {
   case GPif:
-    nb=genlistargs(n,arg,2,STACKSZ);
+    nb=genlistargs(n,arg,2,3);
     gentype(arg[0]);
     gentype(arg[1]);
     tree[n].m=(1<<Melse)|((tree[arg[0]].m|(tree[arg[1]].m&~(1<<Mterm)))&MODHERIT);
-    if (nb>=3)
+    if (nb==3)
     {
-      int i;
-      int t=tree[arg[1]].t, m =tree[arg[1]].m;
-      for(i=2; i<nb;i++)
-      {
-        gentype(arg[i]);
-        m|=tree[arg[i]].m&MODHERIT;
-        t=typemax[t][tree[arg[i]].t];
-      }
-      tree[n].m|=m;
-      return t;
+      gentype(arg[2]);
+      /*We must not set Mterm if we are not sure.
+        In this case we must set Msidef instead.*/
+      tree[n].m|=((tree[arg[2]].m&(~(1<<Mterm)))&MODHERIT);
+      if ((tree[arg[1]].m&(1<<Mterm)) && (tree[arg[2]].m&(1<<Mterm)))
+        tree[n].m|=(1<<Mterm);
+      else
+        if ((tree[arg[1]].m&(1<<Mterm)) || (tree[arg[2]].m&(1<<Mterm)))
+          tree[n].m|=(1<<Msidef);
+      return typemax[tree[arg[1]].t][tree[arg[2]].t];
     }
+    if ((tree[arg[1]].m&(1<<Mterm)))
+      tree[n].m|=(1<<Msidef);
     return tree[arg[1]].t;
   case GPwhilenull:
     nb=genlistargs(n,arg,2,2);
@@ -994,7 +1206,7 @@ int gentypefuncspec(int n, gpfunc *gp)
     return Gvoid;
   case GPforstep:
   case GPforstepvec:
-    nb=genlistargs(n,arg,4,7);
+    nb=genlistargs(n,arg,4,8);
     gentypemode(nb,arg,&tree[n].m);
     tree[n].m&=~(1<<Mterm);
     return Gvoid;
@@ -1124,25 +1336,17 @@ void genentryspec(FILE *fout, int n, gpfunc *gp)
         fprintf(fout,gp->spec==GPbreak?"break":"continue");
       break;
     case GPif:
-      nb=genlistargs(n,arg,2,STACKSZ);
+      nb=genlistargs(n,arg,2,3);
       genindentline(fout,"if (");
       gencast(fout,arg[0],Gbool);
       fprintf(fout,")\n");
-      genbrace(fout,arg[1]);
-      if (nb>=3)
+      if (nb==2)
+        genbrace(fout,arg[1]);
+      else
       {
-        for(i=2;i<nb-1;i+=2)
-        {
-          genindentline(fout,"else if (");
-          gencast(fout,arg[i],Gbool);
-          fprintf(fout,")\n");
-          genbrace(fout,arg[i+1]);
-        }
-        if (i<nb)
-        {
-          genindentline(fout,"else\n");
-          genbrace(fout,arg[i]);
-        }
+        genbrace(fout,arg[1]);
+        genindentline(fout,"else\n");
+        genbrace(fout,arg[2]);
       }
       break;
     case GPiferr:
@@ -1301,21 +1505,25 @@ void genentryspec(FILE *fout, int n, gpfunc *gp)
     }
 #endif
   case GPforstepvec:
-    nb=genlistargs(n,arg,6,6);
+    nb=genlistargs(n,arg,8,8);
     genindentline(fout,"for (");
-    gencode(fout,arg[4]);
+    gencode(fout,arg[6]);
     fprintf(fout," = 0, ");
     gencode(fout,arg[0]);
     fprintf(fout,"; ");
     gencode(fout,arg[3]);
+    fprintf(fout,"?");
+    gencode(fout,arg[4]);
+    fprintf(fout,":");
+    gencode(fout,arg[5]);
     fprintf(fout,"; ");
-    gencode(fout,arg[4]);
+    gencode(fout,arg[6]);
     fprintf(fout," = ");
-    gencode(fout,arg[4]);
+    gencode(fout,arg[6]);
     fprintf(fout," == ");
-    gencast(fout,arg[5],Gsmall);
+    gencast(fout,arg[7],Gsmall);
     fprintf(fout," ? 1 :");
-    gencode(fout,arg[4]);
+    gencode(fout,arg[6]);
     fprintf(fout," + 1, ");
     gencode(fout,arg[2]);
     fprintf(fout,")\n");
@@ -1334,8 +1542,8 @@ void genentryspec(FILE *fout, int n, gpfunc *gp)
     }
     break;
   case GPgpmany:
-    nb=genlistargs(n,arg,3,STACKSZ);
-    bl=block+tree[arg[2]].x;
+    nb=genlistargs(n,arg,2,3);
+    bl=block+tree[arg[1]].x;
     if (!(bl->gc&(1<<GCneeded)))
       break;
     genindent(fout);
@@ -1360,11 +1568,19 @@ void genentryspec(FILE *fout, int n, gpfunc *gp)
     {
       int parg[STACKSZ+1];
       /*FIXME:Putting an if statement in the wilderness is not secure*/
-      if (arg[1]!=GNOARG)
+      if (nb==3)
       {
-        parg[0]=arg[1]; parg[1]=arg[0];
         fprintf(fout,"if (");
-        genfuncbydesc(fout,2,parg,FC_low_stack_lim,n);
+        if (arg[2]==GNOARG)
+        {
+          parg[0]=arg[0];
+          genfuncbydesc(fout,1,parg,FC_gc_needed,n);
+        }
+        else
+        {
+          parg[0]=arg[2]; parg[1]=arg[0];
+          genfuncbydesc(fout,2,parg,FC_low_stack_lim,n);
+        }
         fprintf(fout,")\n");
         indent++;
         genindent(fout);
@@ -1377,6 +1593,11 @@ void genentryspec(FILE *fout, int n, gpfunc *gp)
       fprintf(fout,";\n");
     }
     break;
+  case GPlocalprec:
+  case GPlocalbitprec:
+    gencode(fout, y);
+    preclevel++;
+    break;
   case GPreturn:
     fprintf(fout,"return");
     if (functype(lfunc[currfunc])!=Gvoid)
