@@ -1,5 +1,5 @@
 /*
-Copyright (C) 2000-2005  The PARI group.
+Copyright (C) 2000-2013  The PARI group.
 
 This file is part of the GP2C package.
 
@@ -10,7 +10,7 @@ ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
 with the package; see the file 'COPYING'. If not, write to the Free Software
-Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */
+Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.*/
 
 #include "config.h"
 #include <stdlib.h>
@@ -219,9 +219,10 @@ void makeblocks3(int bl1, int bl2, int bl3, int n, int bseq, int aseq, int ret,
   restorectx(sav0);
 }
 
-static void affectval(int vn, int val, int *seq)
+void affectval(int vn, int val, int *seq)
 {
   ctxvar *v = ctxstack+vn;
+  if (val<0) die(-1,"Internal error: affectval");
   int valn=newnode(Fassign,newleaf(v->node),val);
   *seq=addseqright(*seq,valn);
 }
@@ -251,7 +252,11 @@ int genblockdeclaration(int args, int n, int flag, int type, int *seq)
         /*Make sure GEN objects are gerepilable.*/
         val = newnode(Ftag, newnode(Ftag, newsmall(0), Ggen), tv);
       if (decl==global)
-        fillvar(var,flag,tv,val);
+      {
+        int f=fillvar(var,flag,tv,-1);
+        if (autogc && ctype[tv]==Vgen)
+          affectval(f,val,seq);
+      }
       else
         pushvar(var,flag,tv,val);
       break;
@@ -270,10 +275,12 @@ int genblockdeclaration(int args, int n, int flag, int type, int *seq)
           val=newcall("_const_quote",newstringnode(entryname(stack[i]),-1));
           affectval(fillvar(stack[i],flag,mint,-1),val,seq);
         }
-        else if (autogc)
+        else
         {
+          int f = fillvar(stack[i],flag,mint,-1);
           /*Make sure (implicitly GEN) objects are gerepilable.*/
-          fillvar(stack[i],flag,mint, newsmall(0));
+          if (autogc)
+            affectval(f,  newsmall(0), seq);
         }
         break;
       case function:
@@ -302,7 +309,7 @@ int genblockdeclaration(int args, int n, int flag, int type, int *seq)
             int j, ret;
             int nb=listtostack(tree[x].x,Fmatrixelts,vars,STACKSZ," declaration",n);
             val = tree[stack[i]].y;
-            newdecl((1<<Cauto)|(1<<Cconst),Gvoid,GNIL,&ret);
+            newdecl((1<<Cauto)|(1<<Cconst),Gempty,-1,&ret);
             *seq = addseqright(*seq, geninsertvar(val,ret));
             for (j=0;j<nb;j++)
             {
@@ -342,6 +349,12 @@ int genblockdeclaration(int args, int n, int flag, int type, int *seq)
         break;
       }
       break;
+    case Fvararg:
+      if (decl==function && i==nb-1)
+        pushvar(tree[stack[i]].x,flag|(1<<Cvararg),Gvec,val);
+      else
+        die(n,"invalid variadic");
+      break;
     default:
       die(n,"Incorrect node %s in function declaration",
           funcname(tree[stack[i]].f));
@@ -358,9 +371,10 @@ int genblockcheckargs(int sav1, int seq, userfunc *ufunc)
   for(i=ufunc->narg-1;i>=0;i--)
   {
     ctxvar *v=ctxstack+sav1+i;
+    gpdescarg *rule;
     int check=-1;
     if (!(v->flag&(1<<Carg))) continue;
-    gpdescarg *rule=descfindrules1(v->node, FC_badtype);
+    rule=descfindrules1(v->node, FC_badtype);
     if (rule && rule->args[0].type==vartype(*v))
     {
       check=newcall("_badtype",newleaf(v->node));
@@ -410,7 +424,7 @@ void genblockdeffunc(int n)
   int seq=tree[n].y;
   const char *name=entryname(funcid);
   int args=tree[funcid].y;
-  int bl1,gltop=-1;
+  int bl1,gltop=-1,level=preclevel;
   int savcf,savnpv;
   int sav1,nodebl1;
   userfunc *ufunc;
@@ -457,8 +471,8 @@ void genblockdeffunc(int n)
     int i;
     for (i=sav1;i<s_ctx.n;i++)
     {
-      if (ctxstack[i].flag&(1<<Cglobal))
-        ctxstack[i].flag&=~(1<<Cglobal);
+      if (ctxstack[i].flag&(1<<Cfunction))
+        ctxstack[i].flag&=~(1<<Cfunction);
     }
   }
   copyctx(sav1,block+bl1);
@@ -468,6 +482,7 @@ void genblockdeffunc(int n)
   /*restore globalvar*/
   currfunc=savcf;
   newprivvar=savnpv;
+  preclevel=level;
 }
 
 int newvectoridx(int var, int ind)
@@ -625,7 +640,7 @@ genblockvecaff(int n)
   int aseq, ret, i;
   int nv=listtostack(tree[x].x,Fmatrixelts,vars,STACKSZ,"[,...,]=",n);
   genblock(y,n);
-  newdecl(1<<Cconst,Gvec,-1,&ret);
+  newdecl((1<<Cauto)|(1<<Cconst),Gempty,-1,&ret);
   aseq = geninsertvar(y,ret);
   for(i=0; i<nv; i++)
   {
@@ -877,7 +892,7 @@ void genblock(int n, int p)
         if (uf->flag&(1<<UFclosure))
         {
           /* We assume the closure contains the variable */
-          pushvar(n,(1<<Cuser)|(1<<Carg)|(1<<Ccompo)|(1<<Cclosed),Ggen,-1);
+          pushvar(n,(1<<Cuser)|(1<<Cfunction)|(1<<Carg)|(1<<Ccompo)|(1<<Cclosed),Ggen,-1);
           uf->narg++;
         }
         else
@@ -887,7 +902,7 @@ void genblock(int n, int p)
           * It is discarded in genblockdeffunc. This can be changed.
           */
           int iv=newquotenode(entryname(n),-1);
-          pushvar(n,(1<<Cuser)|(1<<Cglobal)|(1<<Cundeclared),Ggen,iv);
+          pushvar(n,(1<<Cuser)|(1<<Cfunction)|(1<<Cglobal)|(1<<Cundeclared),Ggen,iv);
           if (warn)
             warning(n,"variable undeclared");
         }
@@ -917,40 +932,20 @@ void genblock(int n, int p)
       tree[n].f=Ffunction;
     }
   case Ffunction:/*fall through*/
-    if (x==OPn)
-    {
-      if (tree[y].f==Fsmall)
-      {
-        tree[n]=tree[y];
-        tree[n].x=-tree[n].x;
-        break;
-      }
-      else if (tree[y].f==Fconst)
-      {
-        value_t *val = value + tree[y].x;
-        if (val->type==CSTsmall || val->type==CSTsmallreal)
-        {
-          tree[n]=tree[y];
-          val->val.small=-val->val.small;
-          break;
-        }
-      }
-    }
-    else if (x==OPtrans && tree[y].f==Fvec)
+    if (x==OPtrans && tree[y].f==Fvec)
     {
       if (tree[y].x==-1)
       {
         int z=newgetgvec("t_COL");
         tree[n]=tree[z];
-        break;
       }
       else
       {
         tree[n]=tree[y];
         genblockvector(n,"t_COL");
-        break;
       }
     }
+    else
     {
       int nf=findfunction(value[x].val.str);
       gpfunc *gp = lfunc+currfunc;
@@ -958,7 +953,7 @@ void genblock(int n, int p)
       int nv = getvar(n);
       if (nf<0 && uf->flag&(1<<UFclosure) && nv==-1)
       {
-        pushvar(n,(1<<Cuser)|(1<<Carg)|(1<<Ccompo)|(1<<Cclosed),Ggen,-1);
+        pushvar(n,(1<<Cuser)|(1<<Cfunction)|(1<<Carg)|(1<<Ccompo)|(1<<Cclosed),Ggen,-1);
         uf->narg++;
         genblock(y,n);
       }
@@ -968,6 +963,8 @@ void genblock(int n, int p)
         tree[n]=tree[seq];
         genblock(n,p);
       }
+      else if (nf>=0 && lfunc[nf].iter)
+        genblockfunciter(n,lfunc+nf);
       else if (nf>=0 && lfunc[nf].spec>0)
         genblockfuncspec(n,p,lfunc+nf);
       else if (nf>=0 && lfunc[nf].proto.code)
